@@ -15,6 +15,7 @@
 import copy
 import json
 import re
+import pathlib
 
 from typing import List, Optional, Set
 
@@ -22,8 +23,11 @@ from kfp_tekton.compiler._k8s_helper import sanitize_k8s_name
 from kfp_tekton.compiler._op_to_template import _get_base_step, _add_mount_path, _prepend_steps
 from os import environ as env
 
+BIG_DATA_MIDPATH = "artifacts/$ORIG_PR_NAME"
+BIG_DATA_PATH_FORMAT = "/".join(["$(workspaces.$TASK_NAME.path)", BIG_DATA_MIDPATH, "$TASKRUN_NAME", "$TASK_PARAM_NAME"])
 
-def fix_big_data_passing(workflow: dict) -> dict:
+
+def fix_big_data_passing(workflow: dict, loops_pipeline: dict, loop_name_prefix: str) -> dict:
     """
     fix_big_data_passing converts a workflow where some artifact data is passed
     as parameters and converts it to a workflow where this data is passed as
@@ -92,7 +96,7 @@ def fix_big_data_passing(workflow: dict) -> dict:
 
     pipeline_template = workflow["spec"]["pipelineSpec"]
 
-    pipelinerun_template = workflow
+    pipelinerun_template = load_annotations(workflow)
 
     # 1. Index the pipelines to understand how data is being passed and which
     #  inputs/outputs are connected to each other.
@@ -300,10 +304,13 @@ def fix_big_data_passing(workflow: dict) -> dict:
 
     # Use workspaces to tasks if big data passing instead of 'results', 'copy-inputs'
     for task_template in container_templates:
-        task_template = big_data_passing_tasks(task_template,
+        task_template = big_data_passing_tasks(pipelinerun_name,
+                                               task_template,
                                                pipelinerun_template,
                                                inputs_consumed_as_artifacts,
-                                               outputs_consumed_as_artifacts)
+                                               outputs_consumed_as_artifacts,
+                                               loops_pipeline,
+                                               loop_name_prefix)
 
     # Remove input parameters unless they're used downstream.
     # This also removes unused container template inputs if any.
@@ -315,6 +322,7 @@ def fix_big_data_passing(workflow: dict) -> dict:
                     'name'
                 ),  # TODO: pipeline has no name, use pipelineRun name?
                 input_parameter['name']) in inputs_consumed_as_parameters
+                or input_parameter['name'].endswith("-trname")
         ]
 
     # Remove output parameters unless they're used downstream
@@ -352,6 +360,7 @@ def fix_big_data_passing(workflow: dict) -> dict:
                 (task['name'],
                 parameter_argument['name']) not in inputs_consumed_as_artifacts
                 or task['name'] in resource_template_names
+                or parameter_argument['name'].endswith("-trname")
             ]
 
         # tekton results doesn't support underscore
@@ -366,10 +375,10 @@ def fix_big_data_passing(workflow: dict) -> dict:
                     argument_placeholder_parts[2],
                     sanitize_k8s_name(argument_placeholder_parts[3]))
 
+    workflow = jsonify_annotations(workflow)
     # Need to confirm:
     # I didn't find the use cases to support workflow parameter consumed as artifacts downstream in tekton.
     # Whether this case need to be supporting?
-
     clean_up_empty_workflow_structures(workflow)
     return workflow
 
@@ -414,7 +423,7 @@ def big_data_passing_pipeline(name: str, template: dict, inputs_tasks: set(),
             input_name = parameter_argument['name']
             if (task.get('name'), input_name) in inputs_tasks:
                 pipeline_workspaces.add(pipeline_name)
-                # Add workspaces instead of parmas, for tasks of big data inputs
+                # Add workspaces instead of params, for tasks of big data inputs
                 if not task.setdefault('workspaces', []):
                     task['workspaces'].append({
                         "name": task.get('name'),
@@ -453,10 +462,12 @@ def big_data_passing_pipelinerun(name: str, pr: dict, pw: set):
         # User could modify default size of storage in the yaml file mannually if necessary.
         DEFAULT_ACCESSMODES = env.get('DEFAULT_ACCESSMODES', 'ReadWriteMany')
         DEFAULT_STORAGE_SIZE = env.get('DEFAULT_STORAGE_SIZE', '2Gi')
+        DEFAULT_STORAGE_CLASS = env.get('DEFAULT_STORAGE_CLASS', 'kfp-csi-s3')
         pr['spec']['workspaces'].append({
             "name": pipelinerun_name,
             "volumeClaimTemplate": {
                 "spec": {
+                    "storageClassName": DEFAULT_STORAGE_CLASS,
                     "accessModes": [DEFAULT_ACCESSMODES],
                     "resources": {
                         "requests": {
@@ -470,38 +481,45 @@ def big_data_passing_pipelinerun(name: str, pr: dict, pw: set):
     return pr, prw
 
 
-def big_data_passing_tasks(task: dict, pipelinerun_template: dict,
-                           inputs_tasks: set, outputs_tasks: set) -> dict:
+def big_data_passing_tasks(prname: str, task: dict, pipelinerun_template: dict,
+                            inputs_tasks: set, outputs_tasks: set, loops_pipeline: dict,
+                            loop_name_prefix: str) -> dict:
     task_name = task.get('name')
     task_spec = task.get('taskSpec', {})
     # Data passing for the task outputs
-    task_outputs = task_spec.get('results', [])
-    for task_output in task_outputs:
+    appended_taskrun_name = False
+    for task_output in task.get('taskSpec', {}).get('results', []):
         if (task_name, task_output.get('name')) in outputs_tasks:
-            if not task_spec.setdefault('workspaces', []):
-                task_spec['workspaces'].append({"name": task_name})
+            if not task.get('taskSpec', {}).setdefault('workspaces', []):
+                task.get('taskSpec', {})['workspaces'].append({"name": task_name})
             # Replace the args for the outputs in the task_spec
             # $(results.task_output.get('name').path)  -->
             # $(workspaces.task_name.path)/task_name-task_output.get('name')
             placeholder = '$(results.%s.path)' % (sanitize_k8s_name(
                 task_output.get('name')))
-            workspaces_parameter = '$(workspaces.%s.path)/%s-%s' % (
-                task_name, task_name, task_output.get('name'))
+            workspaces_parameter = '$(workspaces.%s.path)/%s/%s/%s' % (
+                task_name, BIG_DATA_MIDPATH, "$(context.taskRun.name)", task_output.get('name'))
+            # For child nodes to know the taskrun name, it has to pass to results via /tekton/results emptydir
+            if not appended_taskrun_name:
+                copy_taskrun_name_step = _get_base_step('output-taskrun-name')
+                copy_taskrun_name_step['script'] += 'echo -n "%s" > $(results.taskrun-name.path)\n' % ("$(context.taskRun.name)")
+                task['taskSpec']['results'].append({"name": "taskrun-name"})
+                task['taskSpec']['steps'].append(copy_taskrun_name_step)
+                _append_original_pr_name_env(task)
+                appended_taskrun_name = True
             task['taskSpec'] = replace_big_data_placeholder(
                 task.get("taskSpec", {}), placeholder, workspaces_parameter)
-            pipelinerun_template['metadata']['annotations'] = replace_big_data_placeholder(
-                pipelinerun_template['metadata']['annotations'], placeholder, workspaces_parameter)
+            artifact_items = pipelinerun_template['metadata']['annotations']['tekton.dev/artifact_items']
+            artifact_items[task['name']] = replace_big_data_placeholder(
+                artifact_items[task['name']], placeholder, workspaces_parameter)
+            pipelinerun_template['metadata']['annotations']['tekton.dev/artifact_items'] = \
+                artifact_items
 
-    # Remove artifacts outputs from results
-    task.get("taskSpec", {})['results'] = [
-        result for result in task_outputs
-        if (task_name, result.get('name')) not in outputs_tasks
-    ]
-
-    # Data passing for task inputs
     task_spec = task.get('taskSpec', {})
     task_params = task_spec.get('params', [])
     task_artifacts = task_spec.get('artifacts', [])
+
+    # Data passing for task inputs
     for task_param in task_params:
         if (task_name, task_param.get('name')) in inputs_tasks:
             if not task_spec.setdefault('workspaces', []):
@@ -513,8 +531,48 @@ def big_data_passing_tasks(task: dict, pipelinerun_template: dict,
             for task_artifact in task_artifacts:
                 if task_artifact.get('name') == task_param.get('name'):
                     placeholder = task_artifact.get('path')
-            workspaces_parameter = '$(workspaces.%s.path)/%s' % (
-                task_name, task_param.get('name'))
+            task_param_task_name = ""
+            task_param_param_name = ""
+            for o_task in outputs_tasks:
+                if '-'.join(o_task) == task_param.get('name'):
+                    task_param_task_name = o_task[0]
+                    task_param_param_name = o_task[1]
+                    break
+            # If the param name is constructed with task_name-param_name,
+            # use the current task_name as the path prefix
+
+            def append_taskrun_params(task_name_append: str):
+                taskrun_param_name = task_name_append + "-trname"
+                inserted_taskrun_param = False
+                for param in task['taskSpec'].get('params', []):
+                    if param.get('name', "") == taskrun_param_name:
+                        inserted_taskrun_param = True
+                        break
+                if not inserted_taskrun_param:
+                    task['taskSpec']['params'].append({"name": taskrun_param_name})
+                    task['params'].append({"name": taskrun_param_name, "value": "$(tasks.%s.results.taskrun-name)" % task_name_append})
+                    parent_task_queue = [task['name']]
+                    while parent_task_queue:
+                        current_task = parent_task_queue.pop(0)
+                        for loop_name, loop_spec in loops_pipeline.items():
+                            # print(loop_name, loop_spec)
+                            if current_task in loop_spec.get('task_list', []):
+                                parent_task_queue.append(loop_name.replace(loop_name_prefix, ""))
+                                loop_param_names = [loop_param['name'] for loop_param in loops_pipeline[loop_name]['spec']['params']]
+                                if task_name_append + '-taskrun-name' in loop_param_names:
+                                    continue
+                                loops_pipeline[loop_name]['spec']['params'].append({'name': task_name_append + '-taskrun-name',
+                                'value': '$(tasks.%s.results.taskrun-name)' % task_name_append})
+
+            if task_param_task_name:
+                workspaces_parameter = '$(workspaces.%s.path)/%s/$(params.%s-trname)/%s' % (
+                    task_name, BIG_DATA_MIDPATH, task_param_task_name, task_param_param_name)
+                if task_param_task_name != task_name:
+                    append_taskrun_params(task_param_task_name)  # need to get taskrun name from parent path
+            else:
+                workspaces_parameter = '$(workspaces.%s.path)/%s/%s/%s' % (
+                    task_name, BIG_DATA_MIDPATH, "$(context.taskRun.name)", task_param.get('name'))
+            _append_original_pr_name_env(task)
             task['taskSpec'] = replace_big_data_placeholder(
                 task_spec, placeholder, workspaces_parameter)
             task_spec = task.get('taskSpec', {})
@@ -524,10 +582,50 @@ def big_data_passing_tasks(task: dict, pipelinerun_template: dict,
             # add input artifact processes
             task = input_artifacts_tasks(task, task_artifact)
 
+        if (prname, task_artifact.get('name')) in inputs_tasks:
+            # add input artifact processes for pipeline parameter
+            if not task_artifact.setdefault('raw', {}):
+                for i in range(len(pipelinerun_template['spec']['params'])):
+                    param_name = pipelinerun_template['spec']['params'][i]['name']
+                    param_value = pipelinerun_template['spec']['params'][i]['value']
+                    if (task_artifact.get('name') == param_name):
+                        task_artifact['raw']['data'] = param_value
+                        task = input_artifacts_tasks_pr_params(task, task_artifact)
+
+    # If a task produces a result and artifact, add a step to copy artifact to results.
+    artifact_items = pipelinerun_template['metadata']['annotations']['tekton.dev/artifact_items']
+    add_copy_results_artifacts_step = False
+    if task.get("taskSpec", {}):
+        if task_spec.get('results', []):
+            copy_results_artifact_step = _get_base_step('copy-results-artifacts')
+            copy_results_artifact_step['onError'] = 'continue'  # supported by v0.27+ of tekton.
+            copy_results_artifact_step['script'] += 'TOTAL_SIZE=0\n'
+            for result in task_spec['results']:
+                if task['name'] in artifact_items:
+                    artifact_i = artifact_items[task['name']]
+                    for index, artifact_tuple in enumerate(artifact_i):
+                        artifact_name, artifact = artifact_tuple
+                        src = artifact
+                        dst = '$(results.%s.path)' % sanitize_k8s_name(result['name'])
+                        if artifact_name == result['name'] and src != dst:
+                            add_copy_results_artifacts_step = True
+                            copy_results_artifact_step['script'] += (
+                                    'ARTIFACT_SIZE=`wc -c %s | awk \'{print $1}\'`\n' % src +
+                                    'TOTAL_SIZE=$( expr $TOTAL_SIZE + $ARTIFACT_SIZE)\n' +
+                                    'touch ' + dst + '\n' +  # create an empty file by default.
+                                    'if [[ $TOTAL_SIZE -lt 3072 ]]; then\n' +
+                                    '  cp ' + src + ' ' + dst + '\n' +
+                                    'fi\n'
+                            )
+            _append_original_pr_name_env_to_step(copy_results_artifact_step)
+            if add_copy_results_artifacts_step:
+                task['taskSpec']['steps'].append(copy_results_artifact_step)
+
     # Remove artifacts parameter from params
     task.get("taskSpec", {})['params'] = [
         param for param in task_spec.get('params', [])
-        if (task_name, param.get('name')) not in inputs_tasks
+        if (task_name, param.get('name')) not in inputs_tasks or
+            param.get('name').endswith("-trname")
     ]
 
     # Remove artifacts from task_spec
@@ -537,6 +635,28 @@ def big_data_passing_tasks(task: dict, pipelinerun_template: dict,
     return task
 
 
+def input_artifacts_tasks_pr_params(template: dict, artifact: dict) -> dict:
+    copy_inputs_step = _get_base_step('copy-inputs')
+    task_name = template.get('name')
+    task_spec = template.get('taskSpec', {})
+    task_params = task_spec.get('params', [])
+    for task_param in task_params:
+        # For pipeline parameter input artifacts, it will never come from another task because pipeline
+        # params are global parameters. Thus, task_name will always be the executing task name.
+        workspaces_parameter = '$(workspaces.%s.path)/%s/%s/%s' % (
+            task_name, BIG_DATA_MIDPATH, "$(context.taskRun.name)", task_param.get('name'))
+        if 'raw' in artifact:
+            copy_inputs_step['script'] += 'mkdir -p %s\n' % pathlib.Path(workspaces_parameter).parent
+            copy_inputs_step['script'] += 'echo -n "%s" > %s\n' % (
+                artifact['raw']['data'], workspaces_parameter)
+        _append_original_pr_name_env(template)
+
+    template['taskSpec']['steps'] = _prepend_steps(
+        [copy_inputs_step], template['taskSpec']['steps'])
+
+    return template
+
+
 def input_artifacts_tasks(template: dict, artifact: dict) -> dict:
     # The input artifacts in KFP is not pulling from s3, it will always be passed as a raw input.
     # Visit https://github.com/kubeflow/pipelines/issues/336 for more details on the implementation.
@@ -581,3 +701,36 @@ def clean_up_empty_workflow_structures(workflow: list):
             del task['artifacts']
     if not workflow['spec']['pipelineSpec'].setdefault('finally', []):
         del workflow['spec']['pipelineSpec']['finally']
+
+
+def load_annotations(template: dict):
+    artifact_items = json.loads(
+        str(template['metadata']['annotations']['tekton.dev/artifact_items']))
+    template['metadata']['annotations']['tekton.dev/artifact_items'] = \
+        artifact_items
+    return template
+
+
+def jsonify_annotations(template: dict):
+    template['metadata']['annotations']['tekton.dev/artifact_items'] = \
+        json.dumps(template['metadata']['annotations']['tekton.dev/artifact_items'])
+    return template
+
+
+def _append_original_pr_name_env_to_step(step):
+    step.setdefault('env', [])
+    has_original_pr_name = False
+    for task_env in step['env']:
+        if task_env['name'] == "ORIG_PR_NAME":
+            has_original_pr_name = True
+    if not has_original_pr_name:
+        step['env'].append({"name": "ORIG_PR_NAME",
+                            "valueFrom":
+                                {"fieldRef":
+                                    {"fieldPath": "metadata.labels['custom.tekton.dev/originalPipelineRun']"}}})
+
+
+def _append_original_pr_name_env(task_template):
+    for step in task_template['taskSpec']['steps']:
+        if step['name'] == 'main':
+            _append_original_pr_name_env_to_step(step)
