@@ -123,7 +123,7 @@ def _handle_tekton_custom_task(custom_task: dict, workflow: dict, recursive_task
     """
     custom_task_crs = []
     task_list = []
-    tasks = workflow['spec']['pipelineSpec']['tasks']
+    tasks = workflow['spec']['pipelineSpec'].get('tasks', [])
     new_tasks = []
     dependencies = []
     # handle dependecies
@@ -223,18 +223,41 @@ def _handle_tekton_custom_task(custom_task: dict, workflow: dict, recursive_task
                     'type': 'string'
                 } for sub_param in custom_task[custom_task_key]['loop_sub_args']])
 
-            # add loop special filed
+            # add loop special fields
             custom_task_cr['kind'] = 'PipelineLoop'
             if custom_task[custom_task_key]['spec'].get('parallelism') is not None:
                 custom_task_cr['spec']['parallelism'] = custom_task[custom_task_key]['spec']['parallelism']
                 # remove from pipeline run spec
                 del custom_task[custom_task_key]['spec']['parallelism']
             custom_task_cr['spec']['iterateParam'] = custom_task[custom_task_key]['loop_args']
+            separator = custom_task[custom_task_key].get('separator')
+            if separator is not None:
+                custom_task_cr['spec']['iterateParamStringSeparator'] = separator
+            if custom_task[custom_task_key].get('start') is not None:
+                start_end_step_keys = ['from', 'to', 'step']
+                custom_task_cr['spec']['pipelineSpec']['params'] = [value for value
+                                                                    in custom_task_cr['spec']['pipelineSpec']['params']
+                                                                    if value['name'] not in start_end_step_keys]
+                custom_task[custom_task_key]['spec']['params'] = [custom_task_param for custom_task_param
+                                                                  in custom_task[custom_task_key]['spec']['params']
+                                                                  if custom_task_param['name'] != custom_task_cr['spec']['iterateParam']]
+
+                custom_task_cr['spec']['iterateNumeric'] = custom_task_cr['spec']['iterateParam']
+                custom_task_cr['spec'].pop('iterateParam')
             for custom_task_param in custom_task[custom_task_key]['spec']['params']:
                 if custom_task_param['name'] != custom_task[custom_task_key]['loop_args'] and '$(tasks.' in custom_task_param['value']:
                     custom_task_cr = json.loads(
                         json.dumps(custom_task_cr).replace(custom_task_param['value'], '$(params.%s)' % custom_task_param['name']))
 
+            # remove separator from CR params
+            if custom_task[custom_task_key].get('separator') is not None:
+                separator_param = custom_task[custom_task_key]['separator']
+                custom_task_cr['spec']['pipelineSpec']['params'] = [
+                    param
+                    for param in custom_task_cr['spec']['pipelineSpec']['params']
+                    if param['name'] != separator_param
+                ]
+
         # need to process task parameters to replace out of scope results
         # because nested graph cannot refer to task results outside of the sub-pipeline.
         custom_task_cr_task_names = [custom_task_cr_task['name'] for custom_task_cr_task in custom_task_cr['spec']['pipelineSpec']['tasks']]
@@ -349,6 +372,16 @@ def _handle_tekton_custom_task(custom_task: dict, workflow: dict, recursive_task
             if task['name'].replace(task_name_prefix, "") not in task_list_trimmed:
                 new_tasks.append(task)
     workflow['spec']['pipelineSpec']['tasks'] = new_tasks
+    # clean up any redundant parameters
+    for i, task in enumerate(workflow['spec']['pipelineSpec']['tasks']):
+        searched_params = []
+        if task.get('params'):
+            new_params_spec = []
+            for workflow_param in task['params']:
+                if workflow_param['name'] not in searched_params:
+                    new_params_spec.append(workflow_param)
+                    searched_params.append(workflow_param['name'])
+            workflow['spec']['pipelineSpec']['tasks'][i]['params'] = new_params_spec
     return custom_task_crs, workflow
 
 
