@@ -331,6 +331,12 @@ class TektonCompiler(Compiler):
         'spec': {},
         'depends': []
       }
+      if hasattr(sub_group, 'separator') and sub_group.separator is not None:
+        self.loops_pipeline[group_name]['separator'] = sub_group.separator.full_name
+      if hasattr(sub_group, 'start') and sub_group.start is not None:
+        self.loops_pipeline[group_name]['start'] = sub_group.start
+        self.loops_pipeline[group_name]['end'] = sub_group.end
+        self.loops_pipeline[group_name]['step'] = sub_group.step
       for subvarName in sub_group.loop_args.referenced_subvar_names:
         if subvarName != '__iter__':
           self.loops_pipeline[group_name]['loop_sub_args'].append(sub_group.loop_args.full_name + '-subvar-' + subvarName)
@@ -368,12 +374,31 @@ class TektonCompiler(Compiler):
           withparam_value = '$(tasks.%s.results.%s)' % (
             sanitize_k8s_name(pipeline_param.op_name),
             param_name)
+
         self.loops_pipeline[group_name]['spec']['params'] = [{
           "name": sub_group.loop_args.full_name,
           "value": withparam_value
         }]
+      elif hasattr(sub_group, 'items_is_string') and sub_group.items_is_string:
+        loop_args_str_value = sub_group.loop_args.to_str_for_task_yaml()
+        self.loops_pipeline[group_name]['spec']['params'] = [{
+          "name": sub_group.loop_args.full_name,
+          "value": loop_args_str_value
+        }]
       else:
         # Need to sanitize the dict keys for consistency.
+        def process_pipelineparam(s):
+          if "{{pipelineparam" in s:
+            pipe_params = re.findall(r"{{pipelineparam:op=([^ \t\n,]*);name=([^ \t\n,]*)}}", s)
+            for pipe_param in pipe_params:
+              if pipe_param[0] == '':
+                s = s.replace("{{pipelineparam:op=%s;name=%s}}" % pipe_param, '$(params.%s)' % pipe_param[1])
+              else:
+                param_name = sanitize_k8s_name(pipe_param[1])
+                s = s.replace("{{pipelineparam:op=%s;name=%s}}" % pipe_param, '$(tasks.%s.results.%s)' % (
+                  sanitize_k8s_name(pipe_param[0]),
+                  param_name))
+          return s
         loop_arg_value = sub_group.loop_args.to_list_for_task_yaml()
         loop_args_str_value = ''
         sanitized_tasks = []
@@ -381,16 +406,68 @@ class TektonCompiler(Compiler):
           for argument_set in loop_arg_value:
             c_dict = {}
             for k, v in argument_set.items():
+              if isinstance(v, dsl.PipelineParam):
+                if v.op_name is None:
+                  v = '$(params.%s)' % v.name
+                else:
+                  param_name = sanitize_k8s_name(v.name)
+                  v = '$(tasks.%s.results.%s)' % (
+                    sanitize_k8s_name(v.op_name),
+                    param_name)
+              else:
+                if isinstance(v, str):
+                  v = process_pipelineparam(v)
               c_dict[sanitize_k8s_name(k, True)] = v
             sanitized_tasks.append(c_dict)
           loop_args_str_value = json.dumps(sanitized_tasks, sort_keys=True)
         else:
+          for i, value in enumerate(loop_arg_value):
+            if isinstance(value, str):
+              loop_arg_value[i] = process_pipelineparam(value)
           loop_args_str_value = json.dumps(loop_arg_value)
 
         self.loops_pipeline[group_name]['spec']['params'] = [{
           "name": sub_group.loop_args.full_name,
           "value": loop_args_str_value
         }]
+
+      if hasattr(sub_group, 'separator') and sub_group.separator is not None:
+        # separator should be added as a parameter
+        sep_param = {
+          "name": sub_group.separator.full_name,
+          "value": sub_group.separator.value
+        }
+        self.loops_pipeline[group_name]['spec']['params'].append(sep_param)
+
+      if hasattr(sub_group, 'start') and sub_group.start is not None:
+        # start, end, step params should be added as a parameter
+        # isinstance(sub_group.start, dsl.PipelineParam)
+        def process_parameter(parameter):
+          parameter_value = str(parameter)
+          if isinstance(parameter, dsl.PipelineParam):
+            if parameter.op_name:
+              parameter_value = '$(tasks.' + parameter.op_name + '.results.' + sanitize_k8s_name(parameter.name) + ')'
+            else:
+              parameter_value = '$(params.' + parameter.name + ')'
+          return parameter_value
+
+        start_param = {
+          "name": 'from',
+          "value": process_parameter(sub_group.start)
+        }
+        self.loops_pipeline[group_name]['spec']['params'].append(start_param)
+        end_param = {
+          "name": 'to',
+          "value": process_parameter(sub_group.end)
+        }
+        self.loops_pipeline[group_name]['spec']['params'].append(end_param)
+        if sub_group.step is not None:
+          step_param = {
+            "name": 'step',
+            "value": process_parameter(sub_group.step)
+          }
+          self.loops_pipeline[group_name]['spec']['params'].append(step_param)
+
       # get other input params
       for input in inputs.keys():
         if input == sub_group.name:
@@ -481,6 +558,74 @@ class TektonCompiler(Compiler):
 
     return templates
 
+  def _get_dependencies(self, pipeline, root_group, op_groups,
+                        opsgroups_groups, opsgroups, condition_params):
+      """Get dependent groups and ops for all ops and groups.
+      Returns:
+        A dict. Key is group/op name, value is a list of dependent groups/ops.
+        The dependencies are calculated in the following way: if op2 depends on op1,
+        and their ancestors are [root, G1, G2, op1] and [root, G1, G3, G4, op2],
+        then G3 is dependent on G2. Basically dependency only exists in the first uncommon
+        ancesters in their ancesters chain. Only sibling groups/ops can have dependencies.
+      """
+      dependencies = defaultdict(set)
+      for op in pipeline.ops.values():
+          upstream_op_names = set()
+          for param in op.inputs + list(condition_params[op.name]):
+              if param.op_name:
+                  upstream_op_names.add(param.op_name)
+          upstream_op_names |= set(op.dependent_names)
+
+          for upstream_op_name in upstream_op_names:
+              # the dependent op could be either a BaseOp or an opsgroup
+              if upstream_op_name in pipeline.ops:
+                  upstream_op = pipeline.ops[upstream_op_name]
+              elif upstream_op_name in opsgroups:
+                  upstream_op = opsgroups[upstream_op_name]
+              else:
+                  raise ValueError('compiler cannot find the ' +
+                                    upstream_op_name)
+
+              upstream_groups, downstream_groups = self._get_uncommon_ancestors(
+                  op_groups, opsgroups_groups, upstream_op, op)
+              # Convert Argo condition DAG dependency into Tekton condition task dependency
+              while len(upstream_groups) > 0 and 'condition-' in upstream_groups[0]:
+                upstream_groups.pop(0)
+              if len(upstream_groups) > 0:
+                dependencies[downstream_groups[0]].add(upstream_groups[0])
+
+      # Generate dependencies based on the recursive opsgroups
+      # TODO: refactor the following codes with the above
+      def _get_dependency_opsgroup(group, dependencies):
+          upstream_op_names = set(
+              [dependency.name for dependency in group.dependencies])
+          if group.recursive_ref:
+              for param in group.inputs + list(condition_params[group.name]):
+                  if param.op_name:
+                      upstream_op_names.add(param.op_name)
+
+          for op_name in upstream_op_names:
+              if op_name in pipeline.ops:
+                  upstream_op = pipeline.ops[op_name]
+              elif op_name in opsgroups:
+                  upstream_op = opsgroups[op_name]
+              else:
+                  raise ValueError('compiler cannot find the ' + op_name)
+              upstream_groups, downstream_groups = \
+                self._get_uncommon_ancestors(op_groups, opsgroups_groups, upstream_op, group)
+              # Convert Argo condition DAG dependency into Tekton condition task dependency
+              while len(upstream_groups) > 0 and 'condition-' in upstream_groups[0]:
+                upstream_groups.pop(0)
+              if len(upstream_groups) > 0:
+                dependencies[downstream_groups[0]].add(upstream_groups[0])
+
+          for subgroup in group.groups:
+              _get_dependency_opsgroup(subgroup, dependencies)
+
+      _get_dependency_opsgroup(root_group, dependencies)
+
+      return dependencies
+
   def _get_inputs_outputs(
           self,
           pipeline,
@@ -1386,13 +1531,13 @@ class TektonCompiler(Compiler):
               # Preserve order of params, required by tests.
               pipeline_loop_crs[i]['spec']['pipelineSpec']['params'] =\
                 sorted(pipeline_loop_crs[i]['spec']['pipelineSpec']['params'], key=lambda kv: (kv['name']))
-            t, e = TektonCompiler._inline_tasks(pipeline_loop_crs[i]['spec']['pipelineSpec']['tasks'],
+            t, e = self._inline_tasks(pipeline_loop_crs[i]['spec']['pipelineSpec']['tasks'],
                                                 pipeline_loop_crs, recursive_tasks_names)
             if e:
               pipeline_loop_crs[i]['spec']['pipelineSpec']['tasks'] = t
               inlined_as_taskSpec.extend(e)
         # Step 2. inline pipeline_loop_crs in the workflow
-        workflow_tasks, e = TektonCompiler._inline_tasks(workflow['spec']['pipelineSpec']['tasks'],
+        workflow_tasks, e = self._inline_tasks(workflow['spec']['pipelineSpec']['tasks'],
                                                          pipeline_loop_crs, recursive_tasks_names)
         inlined_as_taskSpec.extend(e)
         workflow['spec']['pipelineSpec']['tasks'] = workflow_tasks
@@ -1481,8 +1626,7 @@ class TektonCompiler(Compiler):
                                                   "_customtask_cr" + str(i + 1) + '.yaml')
     _validate_workflow(workflow)
 
-  @staticmethod
-  def _inline_tasks(tasks: List[Dict[Text, Any]], crs: List[Dict[Text, Any]], recursive_tasks: List[Text]):
+  def _inline_tasks(self, tasks: List[Dict[Text, Any]], crs: List[Dict[Text, Any]], recursive_tasks: List[Text]):
     """
       Scan all the `tasks` and for each taskRef in `tasks` resolve it in `crs`
        and inline them as taskSpec.
@@ -1509,6 +1653,14 @@ class TektonCompiler(Compiler):
                  'spec': crs[i]['spec']}
               inlined_as_taskSpec.append(cr_ref_name)
               workflow_tasks[j].pop('taskRef')
+      if 'taskSpec' in workflow_tasks[j]:
+        workflow_tasks[j]['taskSpec']['metadata'] = workflow_tasks[j]['taskSpec'].get('metadata', {})
+        task_labels = workflow_tasks[j]['taskSpec']['metadata'].get('labels', {})
+        task_labels['pipelines.kubeflow.org/pipelinename'] = task_labels.get('pipelines.kubeflow.org/pipelinename', '')
+        task_labels['pipelines.kubeflow.org/generation'] = task_labels.get('pipelines.kubeflow.org/generation', '')
+        cache_default = self.pipeline_labels.get('pipelines.kubeflow.org/cache_enabled', 'true')
+        task_labels['pipelines.kubeflow.org/cache_enabled'] = task_labels.get('pipelines.kubeflow.org/cache_enabled', cache_default)
+        workflow_tasks[j]['taskSpec']['metadata']['labels'] = task_labels
     return workflow_tasks, inlined_as_taskSpec
 
 
